(*
    This file is part of the RI::ML project -
    which includes RenderMan Interface 3.2 bindings
    for Standard ML and a RIB file generator.

    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

functor FRenderManRIBOutput() : RENDERMAN_INTERFACE =
  struct
    (* Internal RIB rendering context data *)
    type contextData = {linePos : int ref,
                        outStream : TextIO.outstream,
                        objectHandleCount : CountingInteger.t,
                        lightHandleCount : CountingInteger.t}

    val currentContextHandle = ref ~1
    (* FIXME: This means that there's an arbitrary limit of
              100 rendering contexts and arbitrary limits are
              _always_ evil - so this _should_ have been a dynamic array. *)
    val contexts = Array.tabulate (100, fn _ => NONE)
                     : contextData option Array.array

    fun isUnusedContextF (index, c, (_, false)) =
          (case c of
             NONE =>
               (index, true)
           | SOME(_) =>
               (~1, false))
      | isUnusedContextF (_, _, (foundIndex, true)) =
          (foundIndex, true)

    fun findUnusedContext () =
          #1 (Array.foldli isUnusedContextF (~1, false) (contexts, 0, NONE))

    fun getContextData () =
          if (!currentContextHandle) < 0 orelse
             (!currentContextHandle) >= Array.length contexts then
            NONE
          else
            Array.sub (contexts, !currentContextHandle)

    fun noOutF (s : string) =
          print "ERROR in RIB generator: No RIB file is open! (outF)\n"

    fun noOutNLF () =
          print "ERROR in RIB generator: No RIB file is open! (outNLF)\n"

    fun noOutSPF () =
          print "ERROR in RIB generator: No RIB file is open! (outSPF)\n"

    fun noCountF () =
          (print "ERROR in RIB generator: No RIB file is open! (countF)\n"; 0)
    
    val outFRef = ref noOutF
    val outNLRef = ref noOutNLF
    val outSPRef = ref noOutSPF
    val nextObjectHandleF = ref noCountF
    val nextLightHandleF = ref noCountF

    fun clearContext () =
          (currentContextHandle := ~1;
           outFRef := noOutF;
           outNLRef := noOutNLF;
           outSPRef := noOutSPF;
           nextObjectHandleF := noCountF;
           nextLightHandleF := noCountF)

    val almostMargin = 60

    (* This function is a real exported RenderMan Interface function... *)
    fun Context contextHandle =
          let
            val _ = currentContextHandle := contextHandle
          in
            case getContextData () of
              NONE =>
                clearContext ()
            | SOME({linePos, outStream, objectHandleCount, lightHandleCount}) =>
                let
                  fun outF str =
                        (linePos := (!linePos) + size str;
                         TextIO.output(outStream, str))

                  fun nl () =
                        (outF "\n"; linePos := 0)
                  
                  val _ = outFRef := outF
                  val _ = outNLRef := nl

                  fun checkLineBreak () =
                        if (!linePos) > almostMargin then
                          nl ()
                        else
                          ()

                  fun outSp () =
                        (outF " ";
                         checkLineBreak ())

                  val _ = outSPRef := outSp

                  fun nextObjectHandle () =
                        CountingInteger.next objectHandleCount

                  val _ = nextObjectHandleF := nextObjectHandle

                  fun nextLightHandle () =
                        CountingInteger.next lightHandleCount

                  val _ = nextLightHandleF := nextLightHandle
                in
                  ()
                end
          end

    (* Internal functions *)
    fun outF str =
          (!outFRef) str

    fun nl () =
          (!outNLRef) ()

    fun outSp () =
          (!outSPRef) ()

    fun getLightHandle NONE =
          (!nextLightHandleF) ()
      | getLightHandle (SOME(h)) =
          h

    fun getObjectHandle NONE =
          (!nextObjectHandleF) ()
      | getObjectHandle (SOME(h)) =
          h

    fun outSpList f l =
          List.app (fn e => (outSp (); f e)) l

    fun mapTildeToMinus str =
          String.map (fn c => if c = #"~" then #"-" else c) str

    fun outReal r =
          outF (mapTildeToMinus (Real.toString r))

    val outReals = outSpList outReal

    fun outInt i =
          outF (mapTildeToMinus (Int.toString i))

    val outInts = outSpList outInt

    fun brackets f bracket input =
          (if bracket then outF "[" else ();
           f input;
           if bracket then outF "]" else ())

    fun outPointF {x, y, z} =
          (outReal x;
           outSp (); outReal y;
           outSp (); outReal z)

    fun outPoint bracket p =
          brackets outPointF bracket p

    fun outRGB {r, g, b} =
          (outReal r; outSp ();
           outReal g; outSp ();
           outReal b; outSp ())

    fun outST {s, t} =
          (outReal s; outSp ();
           outReal t; outSp ())

    fun outHpointF {x, y, z, w} =
          (outReal x;
           outSp (); outReal y;
           outSp (); outReal z;
           outSp (); outReal w)

    fun outHpoint bracket p =
          brackets outHpointF bracket p

    fun outColorF value =
          RealArray.app (fn c => (outSp (); outReal c)) value

    fun outColor bracket c =
          brackets outColorF bracket c

    fun outBoundF {xmin, xmax, ymin, ymax, zmin, zmax} =
          (outReal xmin;
           outSp (); outReal xmax;
           outSp (); outReal ymin;
           outSp (); outReal ymax;
           outSp (); outReal zmin;
           outSp (); outReal zmax)

    fun outBound bracket b =
          brackets outBoundF bracket b

    fun outBool b =
          if b then
            outF "1"
          else
            outF "0"

    fun strEscape (c, result) =
          if c = #"\\" then
            result ^ "\\\\"
          else if c = #"\n" then
            result ^ "\\n"
          else if c = #"\r" then
            result ^ "\\r"
          else if c = #"\t" then
            result ^ "\\t"
          else if c = #"\b" then
            result ^ "\\b"
          else if c = #"\f" then
            result ^ "\\f"
          else if c = #"\"" then
            result ^ "\\\""
          (* FIXME: \ddd and \newline are the only escapecodes
                    not implemented here... *)
          else
            result ^ (Char.toString c)

    fun outString s =
          (outF "\"";
           outF (Substring.foldl strEscape "" (Substring.all s));
           outF "\"")

    val outStrings = outSpList outString

    fun outToken token =
          outString (Rt.TokenString token)

    fun outTokens tokens =
          outSpList outToken tokens

    fun outMatrixRow (v0, v1, v2, v3) =
          (outReal v0;
           outSp (); outReal v1;
           outSp (); outReal v2;
           outSp (); outReal v3)

    fun outMatrix m =
          let
            val (r0, r1, r2, r3) = m
          in
            (* The brackets [ and ] seems to be required
               for matrices according to the RenderMan Interface
               specification page 179. *)
            (outF "[";
             outMatrixRow r0; outSp ();
             outMatrixRow r1; outSp ();
             outMatrixRow r2; outSp ();
             outMatrixRow r3;
             outF "]")
          end

    fun outParamValue (Rt.VBoolean value) =
          outBool value
      | outParamValue (Rt.VInt value) =
          outInt value
      | outParamValue (Rt.VFloat value) =
          outReal value
      | outParamValue (Rt.VString value) =
          outString value
      | outParamValue (Rt.VPoint value) =
          outPoint false value
      | outParamValue (Rt.VHpoint value) =
          outHpoint false value
      | outParamValue (Rt.VVector value) =
          outPoint false value
      | outParamValue (Rt.VNormal value) =
          outPoint false value
      | outParamValue (Rt.VColor value) =
          outColor false value
      | outParamValue (Rt.VMatrix value) =
          outMatrix value
      | outParamValue (Rt.VST value) =
          outST value
      | outParamValue (Rt.VLInts value) =
          outInts value
      | outParamValue (Rt.VLFloats value) =
          outReals value
      | outParamValue (Rt.VLStrings value) =
          outStrings value
      | outParamValue (Rt.VLPoints value) =
          outSpList (outPoint false) value
      | outParamValue (Rt.VLHpoints value) =
          outSpList (outHpoint false) value
      | outParamValue (Rt.VLVectors value) =
          outSpList (outPoint false) value
      | outParamValue (Rt.VLNormals value) =
          outSpList (outPoint false) value
      | outParamValue (Rt.VLColors value) =
          outSpList (outColor false) value
      | outParamValue (Rt.VLMatrices value) =
          outSpList outMatrix value
      | outParamValue (Rt.VLSTs value) =
          outSpList outST value

    fun outParam (token, value) =
          (outToken token;
           outSp ();
           outF "[";
           outParamValue value;
           outF "]")

    fun outParams params =
          outSpList outParam params

    (* FIXME: This function supports the same as the format
              function in RenderMan shading language
              (accoding to Advanced RenderMan page 180).
              I'm not quite sure if this is the same that
              the RIB error handling function should support. *)
    fun printfFormatF (#"f", ((Rt.VFloat(value))::params, true)) =
          (outReal value; (params, false))
      | printfFormatF (#"c", ((Rt.VColor(value))::params, true)) =
          (outColor false value; (params, false))
      | printfFormatF (#"p", ((Rt.VPoint(value))::params, true)) =
          (outPoint false value; (params, false))
      | printfFormatF (#"p", ((Rt.VHpoint(value))::params, true)) =
          (outHpoint false value; (params, false))
      | printfFormatF (#"p", ((Rt.VVector(value))::params, true)) =
          (outPoint false value; (params, false))
      | printfFormatF (#"p", ((Rt.VNormal(value))::params, true)) =
          (outPoint false value; (params, false))
      | printfFormatF (#"m", ((Rt.VMatrix(value))::params, true)) =
          (outMatrix value; (params, false))
      | printfFormatF (#"s", ((Rt.VString(value))::params, true)) =
          (outF value; (params, false))
      | printfFormatF (c, (param::params, true)) =
          (outF ("ERROR: Unmatched escape code: %" ^ (str c) ^ "!");
           (params, false))
      | printfFormatF (c, (nil, true)) =
          (outF ("ERROR: No more data for escape code: %" ^ (str c) ^ "!");
           (nil, false))
      | printfFormatF (#"%", (params, false)) =
          (params, true)
      | printfFormatF (c, (params, false)) =
          (outF (str c); (params, false))

    fun outPrintfStyle format params =
          (Substring.foldl printfFormatF (params, false)
                           (Substring.all format);
           ())

    (* General *)
    fun ErrorHandler (errorFunc, errorName) =
          (outF "ErrorHandler";
           outSp (); outToken errorName;
           nl())
          
    fun GetLastError () =
          0 (* FIXME: Error handlers not supported here... *)

    fun Declare (variableName, typeName) =
          (outF "Declare";
           outSp (); outString variableName;
           outSp (); outString typeName;
           nl();
           Rt.MakeToken variableName)

    (* In Adv. RenderMan and RI Spec., not in Companion. *)
    fun ArchiveRecord (commentType, format, params) =
          (outF "#"; (* FIXME: Read appendix E of the Spec. ? *)
           outSp (); outToken commentType; (* FIXME: Just write the comment type? *)
           outSp (); outPrintfStyle format params;
           nl())
          
    (* In Adv. RenderMan and RI Spec. 3.2, not in Companion. *)
    (* FIXME: In RI Spec. a parameterlist is added... *)
    fun ReadArchive (fileName, archiveRecordFunc) =
          (outF "ReadArchive";
           outSp (); outToken fileName;
           nl())

    (* Rendering Contexts *)
    fun GetContext () =
          !currentContextHandle
    
    (* Begin-end block *)
    fun Begin fileName =
          case findUnusedContext () of
            ~1 =>
              clearContext ()
          | index =>
              (Array.update (contexts, index,
                             SOME ({linePos = ref 0,
                                    outStream = TextIO.openOut (Rt.TokenString fileName),
                                    objectHandleCount = CountingInteger.create (),
                                    lightHandleCount = CountingInteger.create ()}));
               Context index;
               outF "version 3.03"; nl())

    fun End () =
          case getContextData () of
            SOME({outStream, ...}) =>
              (TextIO.flushOut outStream;
               TextIO.closeOut outStream;
               Array.update (contexts, !currentContextHandle, NONE);
               clearContext ())
          | NONE =>
              print "No RIB file has been opened! (End)\n"

    (* In Adv. RenderMan, not in Companion or RI Spec. 3.2 *)
    fun Synchronize str =
          (* str must always be "abort" *)
          End ()

    (* Frame block *)
    fun FrameBegin frameNo =
          (outF "FrameBegin ";
           outF (Int.toString frameNo);
           nl())
    fun FrameEnd () =
          (outF "FrameEnd"; nl())

    (* World block *)
    fun WorldBegin () =
          (outF "WorldBegin"; nl())
    fun WorldEnd () =
          (outF "WorldEnd"; nl())

    (* Attribute block *)
    fun AttributeBegin () =
          (outF "AttributeBegin"; nl())
    fun AttributeEnd () =
          (outF "AttributeEnd"; nl())

    (* Transform block *)
    fun TransformBegin () =
          (outF "TransformBegin"; nl())
    fun TransformEnd () =
          (outF "TransformEnd"; nl())

    (* Solid block *)
    fun SolidBegin operation =
          (outF "SolidBegin";
           outSp (); outToken operation;
           nl())
    fun SolidEnd () =
          (outF "SolidEnd"; nl())

    (* Object block *)
    fun ObjectBegin handleOption =
          let
            (* FIXME: The Spec. says that handles must be
                      in the range [0, 65535] *)
            val h = getObjectHandle handleOption
          in
            (outF "ObjectBegin";
             outSp (); outInt h;
             nl();
             h)
          end
    fun ObjectEnd () =
          (outF "ObjectEnd"; nl())

    (* Motion block *)
    fun MotionBegin times =
          (outF "MotionBegin [";
           outReals times;
           outF "]";
           nl())
    fun MotionEnd () =
          (outF "MotionEnd"; nl())
    
    (* Options *)
    fun Format (xRes, yRes, pixelAspectRatio) =
          (outF "Format ";
           outF (Int.toString xRes ^ " ");
           outF (Int.toString yRes ^ " ");
           outF (Real.toString pixelAspectRatio);
           nl())

    fun FrameAspectRatio ratio =
          (outF "FrameAspectRatio ";
           outF (Real.toString ratio);
           nl())

    fun ScreenWindow (xMin, xMax, yMin, yMax) =
          (outF "ScreenWindow ";
           outF (Real.toString xMin ^ " ");
           outF (Real.toString xMax ^ " ");
           outF (Real.toString yMin ^ " ");
           outF (Real.toString yMax);
           nl())

    fun CropWindow (xMin, xMax, yMin, yMax) =
          (outF "CropWindow ";
           outF (Real.toString xMin ^ " ");
           outF (Real.toString xMax ^ " ");
           outF (Real.toString yMin ^ " ");
           outF (Real.toString yMax);
           nl())

    fun Projection (name, params) =
          (outF "Projection";
           outSp (); outToken name;
           outParams params;
           nl())

    fun Clipping (nearPlane, farPlane) =
          (outF "Clipping ";
           outF (Real.toString nearPlane ^ " ");
           outF (Real.toString farPlane);
           nl())

    fun ClippingPlane (x, y, z, nx, ny, nz) =
          (outF "ClippingPlane";
           outSp (); outReal x;
           outSp (); outReal y;
           outSp (); outReal z;
           outSp (); outReal nx;
           outSp (); outReal ny;
           outSp (); outReal nz;
           nl())

    fun DepthOfField (fStop, focalLength, focalDistance) =
          (outF "DepthOfField ";
           outF (Real.toString fStop ^ " ");
           outF (Real.toString focalLength ^ " ");
           outF (Real.toString focalDistance);
           nl())

    fun Shutter (openTime, closeTime) =
          (outF "Shutter ";
           outF (Real.toString openTime ^ " ");
           outF (Real.toString closeTime);
           nl())

    fun PixelVariance variance =
          (outF "PixelVariance ";
           outF (Real.toString variance);
           nl())

    fun PixelSamples (nx, ny) =
          (outF "PixelSamples";
           outSp (); outReal nx;
           outSp (); outReal ny;
           nl())

    fun PixelFilter ((filterFunc, filterFuncName), xWidth, yWidth) =
          (outF "PixelFilter";
           outSp (); outToken filterFuncName;
           outSp (); outReal xWidth;
           outSp (); outReal yWidth;
           nl())

    fun Exposure (gain, gamma) =
          (outF "Exposure";
           outSp (); outReal gain;
           outSp (); outReal gamma;
           nl())

    fun Imager (name, params) =
          (outF "Imager";
           outSp (); outToken name;
           outParams params;
           nl())

    fun Quantize (imageType, one, min, max, dither) =
          (outF "Quantize";
           outSp (); outToken imageType;
           outSp (); outInt one;
           outSp (); outInt min;
           outSp (); outInt max;
           outSp (); outReal dither;
           nl())

    fun Display (name, displayType, displayMode, params) =
          (outF "Display";
           outSp (); outToken name;
           outSp (); outToken displayType;
           outSp (); outToken displayMode;
           outParams params;
           nl())

    fun Hider (hiderType, params) =
          (outF "Hider";
           outSp (); outToken hiderType;
           outParams params;
           nl())

                   (* N, nRGB (Nx3 matrix), RGBn (3xN matrix) *)
    fun ColorSamples (n, nRGB, {r, g, b}) =
          (outF "ColorSamples";
           outSp (); outInt n;
           outSp (); outF ("[");
           List.app outRGB nRGB;
           outF ("]");
           outSp (); outF ("[");
           outReals r;
           outReals g;
           outReals b;
           outF ("]");
           nl())
    
    fun RelativeDetail scale =
          (outF "RelativeDetail ";
           outF (Real.toString scale);
           nl())

    fun Option (name, params) =
          (outF "Option";
           outSp (); outToken name;
           outParams params;
           nl())

    (* Lights *)
    fun LightSource (name, handleOption, params) =
          let
            (* FIXME: The Spec. says that handles must be
                      in the range [0, 65535] *)
            val h = getLightHandle handleOption
          in
            (outF "LightSource";
             outSp (); outToken name;
             outSp (); outInt h;
             outParams params;
             nl();
             h)
          end

    fun AreaLightSource (name, handleOption, params) =
          let
            (* FIXME: The Spec. says that handles must be
                      in the range [0, 65535] *)
            val h = getLightHandle handleOption
          in
            (outF "AreaLightSource";
             outSp (); outToken name;
             outSp (); outInt h;
             outParams params;
             nl();
             h)
          end

    (* Attributes *)
    fun Attribute (name, params) =
          (outF "Attribute";
           outSp (); outToken name;
           outParams params;
           nl())

    fun Color color =
          (outF "Color";
           outSp (); outColor true color;
           nl())

    fun Opacity color =
          (outF "Opacity";
           outSp (); outColor true color;
           nl())

    fun Surface (name, params) =
          (outF "Surface";
           outSp (); outToken name;
           outParams params;
           nl())

    fun Atmosphere (name, params) =
          (outF "Atmosphere";
           outSp (); outToken name;
           outParams params;
           nl())

    fun Interior (name, params) =
          (outF "Interior";
           outSp (); outToken name;
           outParams params;
           nl())

    fun Exterior (name, params) =
          (outF "Exterior";
           outSp (); outToken name;
           outParams params;
           nl())

    fun Displacement (name, params) =
          (outF "Displacement";
           outSp (); outToken name;
           outParams params;
           nl())

    fun Illuminate (lightHandle, state) =
          (outF "Illuminate";
           outSp (); outInt lightHandle;
           outSp (); outBool state;
           nl())

    fun TextureCoordinates (s1, t1, s2, t2, s3, t3, s4, t4) =
          (* FIXME: I hope this is correct - check if it needs [ and ] *)
          (outF "TextureCoordinates ";
           outReal s1; outSp();
           outReal t1; outSp();
           outReal s2; outSp();
           outReal t2; outSp();
           outReal s3; outSp();
           outReal t3; outSp();
           outReal s4; outSp();
           outReal t4;
           nl())

    fun ShadingRate rate =
          (outF "ShadingRate ";
           outReal rate;
           nl())

    fun ShadingInterpolation style =
          (outF "ShadingInterpolation";
           outSp (); outToken style;
           nl())

    fun Matte flag =
          (outF "Matte ";
           outBool flag;
           nl())

    fun Bound value =
          (* FIXME: I hope this is correct - check if it needs [ and ] *)
          (outF "Bound";
           outSp (); outBound false value;
           nl())

    fun Detail value =
          (* FIXME: I hope this is correct - check if it needs [ and ] *)
          (outF "Detail";
           outSp (); outBound false value;
           nl())
    
    fun DetailRange (start, low, high, ending) =
          (* FIXME: I hope this is correct - check if it needs [ and ] *)
          (outF "Detail ";
           outReal start; outSp();
           outReal low; outSp();
           outReal high; outSp();
           outReal ending;
           nl())

    fun GeometricApproximation (typ, value) =
          (outF "GeometricApproximation";
           outSp (); outToken typ;
           outSp (); outReal value;
           nl())

    fun Orientation direction =
          (outF "Orientation";
           outSp (); outToken direction;
           nl())

    fun ReverseOrientation () =
          (outF "ReverseOrientation";
           nl())

    fun Sides n =
          (outF "Sides ";
           outInt n;
           nl())

    fun Basis (uBasis, uStep, vBasis, vStep) =
          (* FIXME: This way we cannot use a string as a matrix,
                    which RenderMan allows! But of course it's
                    only used for the built-in matrices, which
                    we should have as matrix values in the
                    RI bindings anyway... *)
          (outF "Basis ";
           outMatrix uBasis; outSp ();
           outInt uStep; outSp ();
           outMatrix vBasis; outSp ();
           outInt vStep;
           nl())
    
    (* Transforms *)
    fun Identity () =
          (outF "Identity";
           nl())
          
    fun Transform matrix =
          (outF "Transform ";
           outMatrix matrix;
           nl())

    fun ConcatTransform matrix =
          (outF "ConcatTransform";
           outSp (); outMatrix matrix;
           nl())

    fun Perspective fov =
          (outF "Perspective ";
           outReal fov;
           nl())
    
    fun Translate (tx, ty, tz) =
          (outF "Translate ";
           outReal tx; outSp();
           outReal ty; outSp();
           outReal tz;
           nl())
    
    fun Rotate (angle, vx, vy, vz) =
          (outF "Rotate ";
           outReal angle; outSp();
           outReal vx; outSp();
           outReal vy; outSp();
           outReal vz;
           nl())
    
    fun Scale (sx, sy, sz) =
          (outF "Scale ";
           outReal sx; outSp();
           outReal sy; outSp();
           outReal sz;
           nl())

    (* Transformations *)
    fun Skew (angle, vx, vy, vz, ax, ay, az) =
          (* FIXME: I hope this is correct - check if it needs [ and ] *)
          (outF "Skew ";
           outReal angle; outSp();
           outReal vx; outSp();
           outReal vy; outSp();
           outReal vz; outSp();
           outReal ax; outSp();
           outReal ay; outSp();
           outReal az;
           nl())

    fun CoordinateSystem name =
          (outF "CoordinateSystem";
           outSp (); outToken name;
           nl())

    (* In Adv. RenderMan and RI Spec., not in Companion. *)
    fun CoordSysTransform name =
          (outF "CoordSysTransform";
           outSp (); outToken name;
           nl())

    fun TransformPoints (fromSpace, toSpace, points) =
          (* Notice: This is not supported in RIB files,
                     so we should really just DO the transformation! *)
          (* FIXME: Not implemented yet! *)
          points
     
    (* Polygons *)
    fun Polygon params =
          (* Notice that nvertices is not written
             to the RIB file - at least this seems to be
             the case for BMRT polygons. *)
          (outF "Polygon";
           outParams params;
           nl())

    fun GeneralPolygon (loopVerts, params) =
          (outF "GeneralPolygon [";
           outInts loopVerts;
           outF "]";
           outParams params;
           nl())
    
    fun PointsPolygons (nVerts, vertIDs, params) =
          (outF "PointsPolygons [";
           outInts nVerts;
           outF "] [";
           outInts vertIDs;
           outF "]";
           outParams params;
           nl())

    fun PointsGeneralPolygons (nLoops, loopVerts, vertIDs, params) =
          (outF "PointsGeneralPolygons [";
           outInts nLoops;
           outF "] [";
           outInts loopVerts;
           outF "] [";
           outInts vertIDs;
           outF "]";
           outParams params;
           nl())

    (* Experimental polygon functions for increased efficiency in ML. *)

    (* val PointsPolygonsRefList : (Rt.Point ref list list) *
                                (* list of points
                                   (FIXME: should be in Rt.ParameterList
                                   as token "P" *)
                                   Rt.Point ref list *
                                   Rt.Parameterlist -> unit
    *)

    (*
    fun PointsPolygonsList polys =
          let
            (* This does not reuse vertices because there's
               no information to do that from.
               But this implmentation is quite elegant :) *)
            val npolys = length polys
            val nverts = map length polys
            val totalvertices = foldl op+ 0 nverts
            val vertids = List.tabulate (totalvertices, fn i => i)
            val paramP = ("P", Rt.VList (map Rt.VPoint (foldl op@ nil polys)))
          in
            PointsPolygons (npolys, nverts, vertids, [paramP])
          end *)

    fun PointsPolygonsList polys =
          (* An elegant way of generating polygons from PointsPolygonsList *)
          List.app (fn poly =>
                      Polygon [(RI.P, Rt.VLPoints poly)])
                   polys
    
    (* End of experimental functions *)

    (* Patches *)
    fun Patch (patchType, params) =
          (outF "Patch";
           outSp (); outToken patchType;
           outParams params;
           nl())

    fun PatchMesh (patchType, nu, uWrap, nv, vWrap, params) =
          (outF "PatchMesh";
           outSp (); outToken  patchType;
           outSp (); outInt nu;
           outSp (); outToken uWrap;
           outSp (); outInt nv;
           outSp (); outToken vWrap;
           outParams params;
           nl())

    fun NuPatch (nuCVs, uOrder, uKnot, uMin, uMax,
                 nvCVs, vOrder, vKnot, vMin, vMax,
                 params) =
          (outF "NuPatch";
           outSp (); outInt nuCVs;
           outSp (); outInt uOrder;
           outReals uKnot;
           outSp (); outReal uMin;
           outSp (); outReal uMax;
           outSp (); outInt nvCVs;
           outSp (); outInt vOrder;
           outReals vKnot;
           outSp (); outReal vMin;
           outSp (); outReal vMax;
           outParams params;
           nl())

    (* Notice: This function is listed under "attributes"
               in RenderMan Companion, but I think it
               really belongs here... *)
    fun TrimCurve (nCurves, order, knot, min, max, n, u, v, w) =
          (* FIXME: I'm not sure if the [ and ] should be there. *)
          (* FIXME: According to Advanced RenderMan, we need to write
                      the value of  length nCurves  (nLoops) first,
                      but not according to the RI Spec. - so I assume that
                      this is correct... *)
          (outF "TrimCurve";
           outSp (); outF "["; outInts nCurves; outF "]";
           outSp (); outF "["; outInts order; outF "]";
           outSp (); outF "["; outReals knot; outF "]";
           outSp (); outF "["; outReals min; outF "]";
           outSp (); outF "["; outReals max; outF "]";
           outSp (); outF "["; outInts n; outF "]";
           outSp (); outF "["; outReals u; outF "]";
           outSp (); outF "["; outReals v; outF "]";
           outSp (); outF "["; outReals w; outF "]";
           nl())

    (* SubdivisionMesh *)
    fun SubdivisionMesh (scheme, nVerts, vertIDs, tags,
                         nArgs, intArgs, floatArgs,
                         params) =
          (outF "SubdivisionMesh";
           outSp (); outToken scheme;
           outSp (); outF "["; outInts nVerts; outF "]";
           outSp (); outF "["; outInts vertIDs; outF "]";
           outSp (); outF "["; outTokens tags; outF "]";
           outSp (); outF "["; outInts nArgs; outF "]";
           outSp (); outF "["; outInts intArgs; outF "]";
           outSp (); outF "["; outReals floatArgs; outF "]";
           outParams params;
           nl())
    
    (* Lightweight primitives *)
    fun Points params =
          (outF "Points";
           outParams params;
           nl())

    fun Curves (curveType, nVertices, wrap, params) =
          (outF "Curves";
           outSp (); outToken curveType;
           outInts nVertices;
           outSp (); outToken wrap;
           outParams params;
           nl())

    fun Blobby (nleaf, codes, reals, strings, params) =
          (outF "Blobby";
           outSp (); outInt nleaf;
           outInts codes;
           outReals reals;
           outStrings strings;
           outParams params;
           nl())
    
    (* Quadrics *)
    fun Sphere (radius, zMin, zMax, thetaMax, params) =
          (outF "Sphere ";
           outReal radius; outSp ();
           outReal zMin; outSp ();
           outReal zMax; outSp ();
           outReal thetaMax;
           outParams params;
           nl())

    fun Cone (height, radius, sweepAngle, params) =
          (outF "Cone ";
           outReal height; outSp ();
           outReal radius; outSp ();
           outReal sweepAngle;
           outParams params;
           nl())

    fun Cylinder (radius, zMin, zMax, sweepAngle, params) =
          (outF "Cylinder ";
           outReal radius; outSp ();
           outReal zMin; outSp ();
           outReal zMax; outSp ();
           outReal sweepAngle;
           outParams params;
           nl())

    fun Hyperboloid (point1, point2, sweepAngle, params) =
          (outF "Hyperboloid ";
           outPoint true point1; outSp ();
           outPoint true point2; outSp ();
           outReal sweepAngle;
           outParams params;
           nl())

    fun Paraboloid (topRadius, zMin, zMax, sweepAngle, params) =
          (outF "Paraboloid ";
           outReal topRadius; outSp ();
           outReal zMin; outSp ();
           outReal zMax; outSp ();
           outReal sweepAngle;
           outParams params;
           nl())

    fun Disk (height, radius, sweepAngle, params) =
          (outF "Disk ";
           outReal height; outSp ();
           outReal radius; outSp ();
           outReal sweepAngle;
           outParams params;
           nl())

    fun Torus (majorRadius, minorRadius, phiMin, phiMax, sweepAngle, params) =
          (outF "Torus ";
           outReal majorRadius; outSp ();
           outReal minorRadius; outSp ();
           outReal phiMin; outSp ();
           outReal phiMax; outSp ();
           outReal sweepAngle;
           outParams params;
           nl())
  
    (* General objects *)
    fun ObjectInstance objectHandle =
          (outF "ObjectInstance";
           outSp (); outInt objectHandle;
           nl())

    fun Procedural (Rt.BlStrings blindData,
                    boundingBox,
                    (subdivideF, subdivideName), free) =
          (outF "Procedural";
           outSp (); outToken subdivideName;
           outSp (); outF "["; outStrings blindData; outF "]";
           outSp (); outBound false boundingBox;
           nl())
      | Procedural (_, _, _, _) =
          () (* Only Delayed ReadArchive, Run Program and Dynamic Load
                is supported when exporting to RIB, and data to
                these functions must be packed in Rt.BlStrings. *)

    fun Geometry (geomType, params) =
          (outF "Geometry";
           outSp (); outToken geomType;
           outParams params;
           nl())
          
    (* Map making *)
    fun MakeTexture (imageName, textureName, sWrap, tWrap,
                     (filterF, filterName), sWidth, tWidth, params) =
          (outF "MakeTexture";
           outSp (); outString imageName;
           outSp (); outString textureName;
           outSp (); outToken sWrap;
           outSp (); outToken tWrap;
           outSp (); outToken filterName;
           outSp (); outReal sWidth;
           outSp (); outReal tWidth;
           outParams params;
           nl())

    fun MakeLatLongEnvironment (imageName, textureName,
                                (filterF, filterName), sWidth, tWidth, params) =
          (outF "MakeLatLongEnvironment";
           outSp (); outString imageName;
           outSp (); outString textureName;
           outSp (); outToken filterName;
           outSp (); outReal sWidth;
           outSp (); outReal tWidth;
           outParams params;
           nl())

    fun MakeCubeFaceEnvironment (pxImage, nxImage,
                                 pyImage, nyImage,
                                 pzImage, nzImage,
                                 textureName, fov,
                                 (filterF, filterName),
                                 sWidth, tWidth, params) =
          (outF "MakeCubeFaceEnvironment";
           outSp (); outString pxImage;
           outSp (); outString nxImage;
           outSp (); outString pyImage;
           outSp (); outString nyImage;
           outSp (); outString pzImage;
           outSp (); outString nzImage;
           outSp (); outString textureName;
           outSp (); outReal fov;
           outSp (); outToken filterName;
           outSp (); outReal sWidth;
           outSp (); outReal tWidth;
           outParams params;
           nl())

    fun MakeShadow (imageName, textureName, params) =
          (outF "MakeShadow";
           outSp (); outString imageName;
           outSp (); outString textureName;
           outParams params;
           nl())

    (* The rest is actually constants, but they are prefixed
       with Ri in the RenderMan Interface and contains mixed case letters -
       so they belong here. They have been implemented in a separate
       module (RIFCs) so that RenderMan implementors don't
       have to reimplement it. *)
    open RIFCs
  end

structure RIB = FRenderManRIBOutput()