(*
    This file is part of RI::ML - the Standard ML bindings for
    the RenderMan Interface version 3.2.

    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    The RenderMan (R) Interface Procedures and Protocol are:
    Copyright 1988, 1989, 2000, Pixar
    All Rights Reserved

    RenderMan (R) is a registered trademark of Pixar

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* All the Rt prefixes have been removed since they are not
   needed when the types are grouped in a structure. *)
functor FRenderManTypes() (* :> RENDERMAN_TYPES *) =
  struct
    (* RenderMan types *)
    
    type Point = {x : real, y : real, z : real}
    type Hpoint = {x : real, y : real, z : real, w : real}
    type Vector = Point
    type Normal = Point
    type Matrix = (real * real * real * real) *
                  (real * real * real * real) *
                  (real * real * real * real) *
                  (real * real * real * real)
    type Bound = {xmin : real, xmax : real,
                  ymin : real, ymax : real,
                  zmin : real, zmax : real}
    type Basis = Matrix
    type Color = RealArray.array

    (* Handle types *)
    type ObjectHandle = int
    type LightHandle = int
    type ContextHandle = int

    (* Tokens *)
    (* The Token type has been made opaque
       to emphasize on typesafety.
       It should be advantageous to have a unique integer
       for each token - and the RenderMan spec. encourages this! *)
    (* Notice: This could be re-implemented so that we don't carry around
               the string for a token - thus saving memory. However,
               this would make the function TokenString a little slower. *)
    type Token = int * string

    val tokenCounter = CountingInteger.create ()
    val tokenDictionary = ref TernaryTree.empty : int TernaryTree.t ref

    fun MakeToken str =
          ((case TernaryTree.lookup (!tokenDictionary) str of
              NONE =>
                let
                  (* FIXME: We might run out of integers! *)
                  val n = CountingInteger.next tokenCounter
                  val _ = tokenDictionary :=
                            TernaryTree.insert (!tokenDictionary) str n
                in
                  n
                end
            | SOME(n) =>
                n),
           str)
    
    fun TokenString (_, str) = str
    fun TokenInt (i, _) = i
    fun EqualTokens ((i1, _), (i2, _)) =
          i1 = i2

    (* Datatype of legal values for parameter lists *)
    datatype Value =
      VBoolean of bool
    | VInt of int
    | VFloat of real
    | VString of string
    | VPoint of Point
    | VHpoint of Hpoint
    | VVector of Vector
    | VNormal of Normal
    | VColor of Color
    | VMatrix of Matrix
    | VST of {s : real, t : real} (* In Adv. RenderMan and RI Spec., not in Companion *)

    (* FIXME: I'm not yet sure if we also need support for arrays here... *)
    | VLInts of int list
    | VLFloats of real list
    | VLStrings of string list
    | VLPoints of Point list
    | VLHpoints of Hpoint list
    | VLVectors of Vector list
    | VLNormals of Normal list
    | VLColors of Color list
    | VLMatrices of Matrix list
    | VLSTs of {s : real, t : real} list (* In Adv. RenderMan and RI Spec., not in Companion *)

    (* Parameterlist *)
    type Parameterlist = (Token * Value) list

    (* Datatype for the Procedural function - to allow using the same
       function for "really blind" data and for string lists to be
       put into RIB files. *)
    datatype 'a Blind =
      BlStrings of string list
    | BlAnything of 'a

    (* We don't use pointers to functions in ML
      (nor references to functions) - we just use functions.
      This also seems to be the case as per RI Spec. 3.2.
      However, in ML we accompany the functions with a Token
      specifying the name of the function. *)
    type FilterFunc = (real * real * real * real -> real) * Token
    type ErrorHandler = (int * int * string -> unit) * Token
    type 'a ProcSubdivFunc = (('a Blind) * real -> unit) * Token
    type 'a ProcFreeFunc = (('a Blind) -> unit) * Token
    type ArchiveCallback = (Token * string * Parameterlist -> unit) * Token
  end

structure Rt = FRenderManTypes()